@@ -11,15 +11,17 @@ lang_spec_t langs[] = {
     { "batch", { "bat", "cmd" } },
     { "cc", { "c", "h", "xs" } },
     { "cfmx", { "cfc", "cfm", "cfml" } },
-    { "clojure", { "clj", "cljs", "cljx" } },
+    { "clojure", { "clj", "cljs", "cljc", "cljx" } },
     { "coffee", { "coffee" } },
     { "cpp", { "cpp", "cc", "C", "cxx", "m", "hpp", "hh", "h", "H", "hxx" } },
     { "csharp", { "cs" } },
     { "css", { "css" } },
     { "delphi", { "pas", "int", "dfm", "nfm", "dof", "dpk", "dproj", "groupproj", "bdsgroup", "bdsproj" } },
+    { "ebuild", { "ebuild", "eclass" } },
     { "elisp", { "el" } },
     { "erlang", { "erl", "hrl" } },
     { "fortran", { "f", "f77", "f90", "f95", "f03", "for", "ftn", "fpp" } },
+    { "fsharp", { "fs", "fsi", "fsx" } },
     { "gettext", { "po", "pot", "mo" } },
     { "go", { "go" } },
     { "groovy", { "groovy", "gtmpl", "gpp", "grunit" } },
@@ -30,7 +32,7 @@ lang_spec_t langs[] = {
     { "ini", { "ini" } },
     { "jade", { "jade" } },
     { "java", { "java", "properties" } },
-    { "js", { "js" } },
+    { "js", { "js", "jsx" } },
     { "json", { "json" } },
     { "jsp", { "jsp", "jspx", "jhtm", "jhtml" } },
     { "less", { "less" } },
@@ -42,7 +44,9 @@ lang_spec_t langs[] = {
     { "markdown", { "markdown", "mdown", "mdwn", "mkdn", "mkd", "md" } },
     { "mason", { "mas", "mhtml", "mpl", "mtxt" } },
     { "matlab", { "m" } },
+    { "mathematica", { "m", "wl" } },
     { "mercury", { "m", "moo" } },
+    { "nim", { "nim" } },
     { "objc", { "m", "h" } },
     { "objcpp", { "mm", "h" } },
     { "ocaml", { "ml", "mli", "mll", "mly" } },
@@ -50,10 +54,13 @@ lang_spec_t langs[] = {
     { "parrot", { "pir", "pasm", "pmc", "ops", "pod", "pg", "tg" } },
     { "perl", { "pl", "pm", "pm6", "pod", "t" } },
     { "php", { "php", "phpt", "php3", "php4", "php5", "phtml" } },
+    { "pike", { "pike", "pmod" } },
     { "plone", { "pt", "cpt", "metadata", "cpy", "py" } },
+    { "puppet", { "pp" } },
     { "python", { "py" } },
     { "rake", { "Rakefiles" } },
     { "rs", { "rs" } },
+    { "r", { "R", "Rmd", "Rnw", "Rtex", "Rrst" } },
     { "ruby", { "rb", "rhtml", "rjs", "rxml", "erb", "rake", "spec" } },
     { "rust", { "rs" } },
     { "salt", { "sls" } },
@@ -68,29 +75,35 @@ lang_spec_t langs[] = {
     { "tcl", { "tcl", "itcl", "itk" } },
     { "tex", { "tex", "cls", "sty" } },
     { "tt", { "tt", "tt2", "ttml" } },
+    { "vala", { "vala", "vapi" } },
     { "vb", { "bas", "cls", "frm", "ctl", "vb", "resx" } },
+    { "velocity", { "vm" } },
     { "verilog", { "v", "vh", "sv" } },
     { "vhdl", { "vhd", "vhdl" } },
     { "vim", { "vim" } },
+    { "wsdl", { "wsdl" } },
+    { "wadl", { "wadl" } },
     { "xml", { "xml", "dtd", "xsl", "xslt", "ent" } },
     { "yaml", { "yaml", "yml" } }
 };
 
-unsigned int get_lang_count() {
+size_t get_lang_count() {
     return sizeof(langs) / sizeof(lang_spec_t);
 }
 
-char *make_lang_regex(const char **extensions) {
+char *make_lang_regex(char *ext_array, size_t num_exts) {
     int regex_capacity = 100;
     char *regex = ag_malloc(regex_capacity);
     int regex_length = 3;
     int subsequent = 0;
-    const char **extension;
+    char *extension;
+    size_t i;
 
     strcpy(regex, "\\.(");
 
-    for (extension = extensions; *extension; ++extension) {
-        int extension_length = strlen(*extension);
+    for (i = 0; i < num_exts; ++i) {
+        extension = ext_array + i * SINGLE_EXT_LEN;
+        int extension_length = strlen(extension);
         while (regex_length + extension_length + 3 + subsequent > regex_capacity) {
             regex_capacity *= 2;
             regex = ag_realloc(regex, regex_capacity);
@@ -100,7 +113,7 @@ char *make_lang_regex(const char **extensions) {
         } else {
             subsequent = 1;
         }
-        strcpy(regex + regex_length, *extension);
+        strcpy(regex + regex_length, extension);
         regex_length += extension_length;
     }
 
@@ -109,3 +122,30 @@ char *make_lang_regex(const char **extensions) {
     regex[regex_length++] = 0;
     return regex;
 }
+
+size_t combine_file_extensions(size_t *extension_index, size_t len, char **exts) {
+    /* Keep it fixed as 100 for the reason that if you have more than 100
+     * file types to search, you'd better search all the files.
+     * */
+    size_t ext_capacity = 100;
+    (*exts) = (char *)ag_malloc(ext_capacity * SINGLE_EXT_LEN);
+    memset((*exts), 0, ext_capacity * SINGLE_EXT_LEN);
+    size_t num_of_extensions = 0;
+
+    size_t i;
+    for (i = 0; i < len; ++i) {
+        size_t j = 0;
+        const char *ext = langs[extension_index[i]].extensions[j];
+        do {
+            if (num_of_extensions == ext_capacity) {
+                break;
+            }
+            char *pos = (*exts) + num_of_extensions * SINGLE_EXT_LEN;
+            strncpy(pos, ext, strlen(ext));
+            ++num_of_extensions;
+            ext = langs[extension_index[i]].extensions[++j];
+        } while (ext);
+    }
+
+    return num_of_extensions;
+}
