@@ -137,23 +137,36 @@ int main(int argc, char **argv) {
                 die("error in pthread_create(): %s", strerror(rv));
             }
 #if defined(HAVE_PTHREAD_SETAFFINITY_NP) && defined(USE_CPU_SET)
-            cpu_set_t cpu_set;
-            CPU_ZERO(&cpu_set);
-            CPU_SET(i % num_cores, &cpu_set);
-            rv = pthread_setaffinity_np(workers[i].thread, sizeof(cpu_set), &cpu_set);
-            if (rv != 0) {
-                die("error in pthread_setaffinity_np(): %s", strerror(rv));
+            if (opts.use_thread_affinity) {
+                cpu_set_t cpu_set;
+                CPU_ZERO(&cpu_set);
+                CPU_SET(i % num_cores, &cpu_set);
+                rv = pthread_setaffinity_np(workers[i].thread, sizeof(cpu_set), &cpu_set);
+                if (rv != 0) {
+                    die("error in pthread_setaffinity_np(): %s", strerror(rv));
+                }
+                log_debug("Thread %i set to CPU %i", i, i);
+            } else {
+                log_debug("Thread affinity disabled.");
             }
-            log_debug("Thread %i set to CPU %i", i, i);
 #else
-            log_debug("No CPU affinity.");
+            log_debug("No CPU affinity support.");
 #endif
         }
         for (i = 0; paths[i] != NULL; i++) {
             log_debug("searching path %s for %s", paths[i], opts.query);
             symhash = NULL;
             ignores *ig = init_ignore(root_ignores, "", 0);
-            search_dir(ig, base_paths[i], paths[i], 0);
+            struct stat s = {.st_dev = 0 };
+#ifndef _WIN32
+            /* The device is ignored if opts.one_dev is false, so it's fine
+             * to leave it at the default 0
+             */
+            if (opts.one_dev && lstat(paths[i], &s) == -1) {
+                log_err("Failed to get device information for path %s. Skipping...", paths[i]);
+            }
+#endif
+            search_dir(ig, base_paths[i], paths[i], 0, s.st_dev);
             cleanup_ignore(ig);
         }
         pthread_mutex_lock(&work_queue_mtx);
